import pytest
import numpy as np

import huracanpy


@pytest.mark.parametrize(
    "filename, kwargs, nvars, ncoords, npoints, ntracks",
    [
        (huracanpy.example_TRACK_file, dict(source="TRACK"), 35, 0, 46, 2),
        (huracanpy.example_TRACK_tilt_file, dict(source="TRACK.tilt"), 3, 1, 46, 2),
        (huracanpy.example_csv_file, dict(), 9, 0, 99, 3),
        (huracanpy.example_parquet_file, dict(), 9, 0, 99, 3),
        (huracanpy.example_TRACK_netcdf_file, dict(), 20, 17, 4580, 86),
        (
            huracanpy.example_TRACK_timestep_file,
            dict(
                source="TRACK",
                track_calendar="timestep",
                initial_date="1940-01-01",
            ),
            38,
            0,
            416,
            19,
        ),
        (huracanpy.example_TE_file, dict(source="tempestextremes"), 8, 0, 210, 8),
        (huracanpy.example_CHAZ_file, dict(), 11, 0, 1078, 20),
        (huracanpy.example_MIT_file, dict(), 10, 1, 2138, 11),
        (huracanpy.example_WiTRACK_file, dict(source="witrack"), 14, 0, 3194, 268),
        (None, dict(source="ibtracs", ibtracs_subset="wmo"), 8, 0, 143287, 4540),
        (None, dict(source="ibtracs", ibtracs_subset="usa"), 10, 0, 121806, 4170),
        (huracanpy.example_old_HURDAT_file, dict(source="ecmwf"), 8, 0, 183, 29),
    ],
)
def test_load(filename, kwargs, nvars, ncoords, npoints, ntracks):
    data = huracanpy.load(filename, **kwargs)

    assert len(data) == nvars
    assert len(data.coords) == ncoords
    assert len(data.time) == npoints
    assert len(data.groupby("track_id")) == ntracks
    assert "record" not in data.coords

    if filename != huracanpy.example_TRACK_tilt_file:
        for name in ["track_id", "time", "lon", "lat"]:
            assert name in data


@pytest.mark.parametrize(
    "filename, source",
    [
        (huracanpy.example_TRACK_file, "TRACK"),
        (huracanpy.example_TRACK_tilt_file, "TRACK.tilt"),
        (huracanpy.example_TRACK_netcdf_file, None),
        (huracanpy.example_csv_file, None),
        (huracanpy.example_parquet_file, None),
        (huracanpy.example_TE_file, "tempestextremes"),
        (huracanpy.example_CHAZ_file, None),
        (huracanpy.example_MIT_file, None),
        (huracanpy.example_WiTRACK_file, "witrack"),
        (None, "ibtracs"),
    ],
)
@pytest.mark.parametrize("extension", ["csv", "nc"])
@pytest.mark.parametrize("muddle", [False, True])
def test_save(filename, source, extension, muddle, tmp_path):
    if extension == "csv" and (
        filename == huracanpy.example_TRACK_tilt_file
        or (filename is not None and filename.split(".")[-1] == "nc")
    ):
        pytest.skip(
            "The netCDF file has multiple dimensions so fails because converting to a"
            " dataframe leads to having rows equal to the product of the dimensions"
            " even though the dimensions cover different variables"
        )
    data = huracanpy.load(filename, source=source)

    # Check that save/load gives the same result when the track_id is not monotonic
    # Caused an issue because they got sorted before
    if muddle:
        data = data.sortby("track_id", ascending=False)

    # Copy the data because save modifies the dataset at the moment
    data_orig = data.copy()
    huracanpy.save(data, str(tmp_path / f"tmp_file.{extension}"))

    # Check that the original data is not modified by the save function
    _assert_dataset_identical(data_orig, data)

    # Reload the data and check it is still the same
    # Saving as netcdf does force sorting by track_id so apply this
    if extension == "nc":
        data = data.sortby("track_id")
    data_reload = huracanpy.load(str(tmp_path / f"tmp_file.{extension}"))
    _assert_dataset_identical(data, data_reload)


def _assert_dataset_identical(ds1, ds2):
    assert len(ds1.variables) == len(ds2.variables)
    assert len(ds1.coords) == len(ds2.coords)
    for var in list(ds1.variables) + list(ds1.coords):
        # Work around for xarray inconsistent loading the data as float or double
        # depending on fill_value and scale_factor
        # np.testing.assert_allclose doesn't work for datetime64, object, or string
        if np.issubdtype(ds1[var].dtype, np.number):
            if ds1[var].dtype != ds2[var].dtype:
                rtol = 1e-6
            else:
                rtol = 0
            np.testing.assert_allclose(
                ds1[var].data.astype(ds2[var].dtype), ds2[var].data, rtol=rtol
            )
        else:
            assert (ds1[var].data == ds2[var].data).all()

    assert len(ds1.attrs) == len(ds2.attrs)
    for attr in ds1.attrs:
        assert ds1.attrs[attr] == ds2.attrs[attr]



import pathlib

import xarray as xr

import huracanpy


data_path = pathlib.Path(__file__).parent / "data_interpolate"


def test_interpolate_time(tracks_csv):
    result = huracanpy.interp_time(tracks_csv, tracks_csv.track_id)

    expected = huracanpy.load(str(data_path / "interpolate_time_result.nc"))

    # cf_role="trajectory_id" automatically added when result was saved
    del expected.track_id.attrs["cf_role"]

    xr.testing.assert_identical(result, expected)



"""
Module with function to compute TC-specific categories
"""

import warnings

import pint
from metpy.xarray import preprocess_and_wrap

from ..info import category
from .._metpy import dequantify_results
from ._conventions import _thresholds


def saffir_simpson_category(wind, convention="Saffir-Simpson", wind_units="m s-1"):
    """
    Function to determine the Saffir-Simpson Hurricane Scale (SSHS) category.

    Parameters
    ----------
    wind : array_like
        10-minutes averaged 10m wind in m/s

    convention : str

    wind_units : str, default="m s-1"
        The units of the input array if they are not already provided by the attributes

    Returns
    -------
    array_like
        The category series.
        You can append it to your tracks by running tracks["sshs"] = get_sshs_cat(tracks.wind)
    """
    return category(
        wind,
        bins=_thresholds[convention]["bins"],
        labels=_thresholds[convention]["labels"],
        variable_units=wind_units,
    )


@dequantify_results
@preprocess_and_wrap(wrap_like="slp")
def pressure_category(slp, convention="Klotzbach", slp_units="hPa"):
    """
    Determine the pressure category according to selected convention.

    Parameters
    ----------
    slp : xarray.DataArray
        Minimum Sea-level Pressure series in hPa
    convention : str
        Name of the classification convention you want to use.
            * Klotzbach (default)
            * Simpson
    slp_units : str, default="hPa"
        The units of the input array if they are not already provided by the attributes

    Returns
    -------
    xarray.DataArray
        The category series.
        You can append it to your tracks by running tracks["cat"] = get_pressure_cat(tracks.slp)

    """
    if not isinstance(slp, pint.Quantity) or slp.unitless:
        if slp.min() > 10000 and slp_units == "hPa":
            warnings.warn(
                "Caution, pressure are likely in Pa, they are being converted to hPa "
                "for categorization. In future specify the units explicitly by passing "
                'slp_units="Pa" to this function or setting '
                'slp.attrs["units"] = "Pa"'
            )
            slp = slp / 100

    return category(
        slp,
        bins=_thresholds[convention]["bins"],
        labels=_thresholds[convention]["labels"],
        variable_units=slp_units,
    )



import pytest

import numpy as np
import xarray as xr

import huracanpy


# %% DataArrayAccessor
def test_nunique():
    data = huracanpy.load(huracanpy.example_csv_file)

    N_tracks = data.track_id.hrcn.nunique()
    assert N_tracks == 3


@pytest.mark.parametrize("call_type", ["get", "add"])
@pytest.mark.parametrize(
    "function, function_args, accessor_name, accessor_function_kwargs",
    [
        (huracanpy.info.hemisphere, ["lat"], "hemisphere", {}),
        (huracanpy.info.basin, ["lon", "lat"], "basin", {}),
        (huracanpy.info.is_land, ["lon", "lat"], "is_land", {}),
        (huracanpy.info.is_ocean, ["lon", "lat"], "is_ocean", {}),
        (huracanpy.info.country, ["lon", "lat"], "country", {}),
        (huracanpy.info.continent, ["lon", "lat"], "continent", {}),
        (huracanpy.tc.ace, ["wind10"], "ace", {"wind_name": "wind10"}),
        (
            huracanpy.tc.ace,
            ["wind10", "track_id"],
            "ace",
            {"wind_name": "wind10", "sum_by": "track_id"},
        ),
        # (huracanpy.tc.pace, ["slp", "wind10"], "pace", {"pressure_name": "slp", "wind_name": "wind10"}),
        (huracanpy.info.season, ["track_id", "lat", "time"], "season", {}),
        (
            huracanpy.tc.saffir_simpson_category,
            ["wind10"],
            "saffir_simpson_category",
            {"wind_name": "wind10"},
        ),
        (
            huracanpy.tc.pressure_category,
            ["slp"],
            "pressure_category",
            {"slp_name": "slp"},
        ),
        (huracanpy.calc.distance, ["lon", "lat", "track_id"], "distance", {}),
        (huracanpy.calc.azimuth, ["lon", "lat", "track_id"], "azimuth", {}),
        (
            huracanpy.calc.translation_speed,
            ["lon", "lat", "time", "track_id"],
            "translation_speed",
            {},
        ),
        (
            huracanpy.calc.delta,
            ["wind10", "track_id"],
            "delta",
            {"var_name": "wind10"},
        ),
        # (
        #     huracanpy.calc.get_rate,
        #     ["wind10", "time", "track_id"],
        #     "rate",
        #     {"var_name": "wind10"},
        # ),
        (
            huracanpy.calc.time_from_genesis,
            ["time", "track_id"],
            "time_from_genesis",
            {},
        ),
        (
            huracanpy.calc.time_from_apex,
            ["time", "track_id", "wind10"],
            "time_from_apex",
            {"intensity_var_name": "wind10"},
        ),
        (
            huracanpy.calc.track_duration,
            ["time", "track_id"],
            "track_duration",
            {},
        ),
        #        (huracanpy.calc.get_freq, ["track_id"], "freq", {}),
        #        (huracanpy.calc.get_tc_days, ["time", "track_id"], "tc_days", {}),
        # (huracanpy.calc.get_gen_vals, ["all", "time", "track_id"], "gen_vals", {}),
        # (
        #     huracanpy.calc.get_apex_vals,
        #     ["all", "wind10", "track_id"],
        #     "apex_vals",
        #     {"varname": "wind10"},
        # ),
    ],
)
def test_accessor_methods_match_functions(
    tracks_csv,
    function,
    function_args,
    accessor_name,
    accessor_function_kwargs,
    call_type,
):
    # Skip functions that only have a "get_" version
    if call_type == "add":
        if accessor_name in [
            "track_duration",
            "freq",
            "tc_days",
            "gen_vals",
            "apex_vals",
        ]:
            pytest.skip(f"Accessor function add_{accessor_name} does not exist")
        elif accessor_name in ["ace"] and "sum_by" in accessor_function_kwargs:
            pytest.skip(f"sum_by not a valid argument for add_{accessor_name}")

    # Call the huracanpy function
    # Get the function arguments as arrays. Use "all" as a wildcard for the full dataset
    function_args = [
        tracks_csv[var] if not var == "all" else tracks_csv for var in function_args
    ]
    result = function(*function_args)

    # Call the accessor method
    result_accessor = getattr(tracks_csv.hrcn, f"{call_type}_{accessor_name}")(
        **accessor_function_kwargs
    )

    # When using the "add_" method a new Dataset is returned with the variable added
    # The naming of the new variable is either simply the function name (minus "add_")
    # or the function name plus the name of the variable specified if it can be applied
    # to different variables
    if call_type == "add":
        varname = accessor_name
        if "var_name" in accessor_function_kwargs:
            varname = f"{varname}_{accessor_function_kwargs['var_name']}"
        result_accessor = result_accessor[varname]

    # Check that the function and method return identical results
    assert type(result) is type(result_accessor), (
        "accessor return type differs from function"
    )
    np.testing.assert_equal(
        np.array(result),
        np.array(result_accessor),
        err_msg="accessor output differs from function output",
    )


# %% DatasetAccessor
# Currently keeping tests here that return more than just a DataArray as the testing is
# less generic
def test_get_methods(tracks_csv):
    """Test get_ accessors output is same as function"""
    data = tracks_csv

    ## - pace
    pace_acc, _ = data.hrcn.get_pace(pressure_name="slp", wind_name="wind10")
    pace_fct, model_fct = huracanpy.tc.pace(data.slp, data.wind10)
    np.testing.assert_array_equal(
        pace_acc, pace_fct, err_msg="accessor output differs from function output"
    )

    ## - time components
    year_acc, month_acc, day_acc, hour_acc = data.hrcn.get_time_components()
    year_fct, month_fct, day_fct, hour_fct = huracanpy.info.time_components(data.time)
    np.testing.assert_array_equal(
        year_acc, year_fct, err_msg="Year component does not match"
    )
    np.testing.assert_array_equal(
        month_acc, month_fct, err_msg="Month component does not match"
    )
    np.testing.assert_array_equal(
        day_acc, day_fct, err_msg="Day component does not match"
    )
    np.testing.assert_array_equal(
        hour_acc, hour_fct, err_msg="Hour component does not match"
    )

    ## - track pace
    pace_acc, _ = data.hrcn.get_pace(wind_name="wind10", sum_by="track_id")
    pace_fct, _ = huracanpy.tc.pace(data.slp, data.wind10, sum_by=data.track_id)
    np.testing.assert_array_equal(
        pace_acc,
        pace_fct,
        "Track PACE accessor output differs from function output",
    )

    ## - Genesis Values
    gen_vals_acc = data.hrcn.get_gen_vals(
        time_name="time",
        track_id_name="track_id",
    )
    gen_vals_fct = huracanpy.calc.gen_vals(data, data.time, data.track_id)
    xr.testing.assert_equal(gen_vals_acc, gen_vals_fct)

    ## - Apex Values
    apex_vals_acc = data.hrcn.get_apex_vals(
        track_id_name="track_id", varname="wind10", stat="max"
    )
    apex_vals_fct = huracanpy.calc.apex_vals(
        data, data.wind10, data.track_id, stat="max"
    )
    xr.testing.assert_equal(apex_vals_acc, apex_vals_fct)


def test_interp_methods():
    data = huracanpy.load(huracanpy.example_csv_file)
    interpolated_data_acc = data.hrcn.interp_time(
        freq="1h", track_id_name="track_id", prog_bar=False
    )
    expected_interpolated_data = huracanpy.interp_time(
        data, data.track_id, freq="1h", prog_bar=False
    )
    np.testing.assert_array_equal(
        interpolated_data_acc.time, expected_interpolated_data.time
    )


def test_accessor_sel_id(tracks_csv):
    result = tracks_csv.hrcn.sel_id(0)
    expected = huracanpy.sel_id(tracks_csv, tracks_csv.track_id, 0)

    xr.testing.assert_identical(result, expected)


def test_accessor_trackswhere(tracks_csv):
    result = tracks_csv.hrcn.trackswhere(lambda x: (x.track_id != 1).any())
    expected = huracanpy.trackswhere(
        tracks_csv, tracks_csv.track_id, lambda x: (x.track_id != 1).any()
    )

    xr.testing.assert_identical(result, expected)



import huracanpy


def test_trackswhere():
    tracks = huracanpy.load(huracanpy.example_csv_file)

    tracks["category"] = huracanpy.tc.pressure_category(tracks.slp, slp_units="Pa")

    tracks_subset = huracanpy.trackswhere(
        tracks, tracks.track_id, lambda track: track.category.max() >= 2
    )

    assert set(tracks_subset.track_id.data) == {0, 2}



import functools

import xarray as xr
import pint


def dequantify_results(original_function):
    @functools.wraps(original_function)
    def wrapped_function(*args, **kwargs):
        result = original_function(*args, **kwargs)

        if isinstance(result, tuple):
            return tuple(_dequantify_result(r) for r in result)
        else:
            return _dequantify_result(result)

    return wrapped_function


def _dequantify_result(result):
    if isinstance(result, xr.DataArray) and isinstance(result.data, pint.Quantity):
        return result.metpy.dequantify()
    else:
        return result



import numpy as np

import huracanpy


def test_match():
    ib = huracanpy.load(source="ibtracs", ibtracs_subset="wmo")
    ref_1996 = ib.where(ib.time.dt.year == 1996, drop=True)
    UZ = huracanpy.load(huracanpy.example_year_file)
    UZ1 = UZ.where(UZ.track_id.isin([1207, 1208, 1210, 1212, 1220, 1238]), drop=True)
    UZ2 = UZ.where(UZ.track_id.isin([1207, 1208, 1209, 1211, 1220, 1238]), drop=True)
    M = huracanpy.assess.match([UZ1, UZ2, ref_1996], ["UZ1", "UZ2", "ib"])

    data1, data2, data3 = UZ1, UZ2, ref_1996
    N1 = len(np.unique(data1.track_id.values))  # Number of tracks in dataset 1
    N2 = len(np.unique(data2.track_id.values))  # Number of tracks in dataset 2
    N3 = len(np.unique(data3.track_id.values))  # Number of tracks in dataset 3

    M_not1 = len(M[M.iloc[:, 0].isna()])
    M_not2 = len(M[M.iloc[:, 1].isna()])
    M_not3 = len(M[M.iloc[:, 2].isna()])
    M_all = len(M[M.isna().sum(axis=1) == 0])

    assert len(M) == 6
    assert (N1, N2, N3) == (6, 6, 118)
    assert (M_not1, M_not2, M_not3, M_all) == (1, 1, 1, 3)


def test_scores():
    ib = huracanpy.load(source="ibtracs", ibtracs_subset="wmo")
    ref_1996 = ib.where(ib.time.dt.year == 1996, drop=True)
    UZ = huracanpy.load(huracanpy.example_year_file)
    M = huracanpy.assess.match([UZ, ref_1996], ["UZ", "ib"])

    POD = huracanpy.assess.pod(M, ref_1996, ref_name="ib")
    FAR = huracanpy.assess.far(M, UZ, detected_name="UZ")

    assert 0.63 < POD < 0.64
    assert 0.14 < FAR < 0.15



"""
Functions to plot track/genesis/whatever density

To compute the density, see huracanpy.diags.track_density
"""

import matplotlib.pyplot as plt
import cartopy.crs as ccrs


def density(
    d,
    contourf_kws=dict(cmap="magma_r", levels=10),
    subplot_kws=dict(projection=ccrs.PlateCarree(180)),
    fig_kws=dict(),
    cbar_kwargs={"label": ""},
):
    """Create a map showing the input density

    Parameters
    ----------
    d : xarray.Dataset
        Density map from :func:`huracanpy.calc.density`
    contourf_kws : dict, optional
        Arguments to be passed to :func:`matplotlib.pyplot.contourf`
    subplot_kws : dict, optional
        Arguments to be passed to :func:`matplotlib.pyplot.subplots`' subplot_kw argument
    fig_kws : dict, optional
        Arguments to be passed to :func:`matplotlib.pyplot.subplots`
    cbar_kwargs: dict, optional
        Arguments to be passed to :func:`matplotlib.pyplot.colorbar`

    Returns
    -------
    tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]
        The figure and axes instances created for the plot

    """
    fig, ax = plt.subplots(subplot_kw=subplot_kws, **fig_kws)
    ax.coastlines()
    ax.gridlines(draw_labels=True)
    d.plot.contourf(
        ax=ax, transform=ccrs.PlateCarree(), cbar_kwargs=cbar_kwargs, **contourf_kws
    )
    return fig, ax



"""
Function to categorise
"""

import warnings

import numpy as np
import pint
import pandas as pd

from metpy.xarray import preprocess_and_wrap
from metpy.units import units


@preprocess_and_wrap(wrap_like="variable")
def category(variable, bins, labels=None, variable_units=None):
    """Calculate a generic category from a variable and a set of thresholds

    >>> huracanpy.info.category(tracks.wind, bins = [0,10,20,30], labels = [1,2,3])

    Parameters
    ----------
    variable : array_like
        The variable to be categorized
    bins : array_like
        Bounds for the categories, including upper and lower bounds
    labels : array_like
        Name of the categories. len(labels) = len(bins) -1
    variable_units : str
        The units of the input variable. Only needs to be specified if they are
        different to the units of the bins and they are not already in the attributes of
        the variable.

    Returns
    -------
    array_like
        The category label for each value in the input variable

    """

    if labels is None:
        warnings.warn(
            "labels not provided, categories will be named from 1 to n in the order of"
            "the provided bins"
        )
        labels = [str(i) for i in range(len(bins) - 1)]

    if not isinstance(variable, pint.Quantity) or variable.unitless:
        if variable_units is None and isinstance(bins, pint.Quantity):
            variable_units = str(bins.units)
        variable = variable * units(variable_units)

    if not isinstance(bins, pint.Quantity) or bins.unitless:
        bins = bins * units(variable_units)

    bins = bins.to(variable.units)

    return np.array(pd.cut(variable, bins, labels=labels))



import pandas as pd


from . import _csv, _TRACK, _netcdf, _tempestextremes, witrack, _old_HURDAT
from . import ibtracs


rename_defaults = dict(
    longitude="lon",
    latitude="lat",
    # Names for MIT netCDF
    n_track="track_id",
    lon_track="lon",
    lat_track="lat",
    # Names for CHAZ netCDF
    stormID="track_id",
    # Names for TRACK netCDF
    TRACK_ID="track_id",
)


def load(
    filename=None,
    source=None,
    variable_names=None,
    rename=dict(),
    units=None,
    baselon=None,
    ibtracs_subset="wmo",
    tempest_extremes_unstructured=False,
    tempest_extremes_header_str="start",
    track_calendar=None,
    **kwargs,
):
    """Load track data

    The optional parameters for different sources of tracks (currently **IBTrACS**,
    **TRACK** and **TempestExtremes**) are named {source}_{parameter} (in lower case),
    e.g. "ibtracs_subset".

    Parameters
    ----------
    filename : str, optional
        The file to be loaded. If `source="ibtracs"`, this is not needed as the data is
        either included in huracanpy or downloaded when called. If the filename is
        provided for an online IBTrACS subset, then the raw downloaded data will be
        saved there.
    source : str, optional
        If the file is not a CSV or NetCDF (identified by the file extension) then the
        source needs to be specified to decide how to load the data

        * **track**
        * **track.tilt**
        * **te**, **tempest**, **tempestextremes**, **uz** (For textual format, not csv)
        * **witrack**
        * **ibtracs**
        * **csv**
        * **netcdf**, **nc** (includes support for CHAZ & MIT-Open file provided appropriate variable_names)
        * **old_hurdat**, **ecmwf**

    variable_names : list of str, optional
          When loading data from an ASCII file (TRACK or TempestExtremes), specify the
          list of variables that have been added to the tracks. This does not include
          variables which are in the track by default (e.g. time/lon/lat/vorticity for
          TRACK). If a variable at multiple levels has been added to the tracks, then a
          variable name for each level needs to be included (e.g. vorticity profiles
          with TRACK). If the variable names are not given, then any additional
          variables will be named variable_n, where n goes from 0 to the number of
          variables

    rename : dict, optional
        A mapping of variable names to rename (for netCDF files only). Defaults are

        * longitude -> lon
        * latitude -> lat

        To override any of these defaults, you can pass the same name, e.g.

        >>> tracks = huracanpy.load(..., rename=dict(longitude="longitude"))

    units : dict, optional
        A mapping of variable names to units

        >>> tracks = huracanpy.load(..., units=dict(vorticity="s**-1", slp="hPa"))

    baselon : scalar, optional
        Force the loaded longitudes into the range (baselon, baselon + 360). e.g.
        (0, 360) or (-180, 180)
    ibtracs_subset : str, default="wmo"
        IBTrACS subset. Two offline versions are available:

        * **wmo**: Data with the wmo_* variables. The data as reported by the WMO agency
          responsible for each basin, so methods are not consistent across basins
        * **usa** or **JTWC**: Data with the usa_* variables. The data as recorded by
          the USA/Joint Typhoon Warning Centre. Methods are consistent across basins,
          but may not be complete.

        To download online data, the subsets are the different filesp rovided by IBTrACS.

        * **ACTIVE**: TCs currently active
        * **ALL**: Entire IBTrACS database
        * Specific basins: **EP**, **NA**, **NI**, **SA**, **SI**, **SP**, **WP**
        * **last3years**: self-explanatory
        * **since1980**: Entire IBTrACS database since 1980 (advent of satellite era,
          considered reliable from then on)

    tempest_extremes_unstructured : bool, default=False
        By default the first two columns in TempestExtremes files are the i, j indices
        of the closest gridpoint, but for unstructured grids it is a single lookup index
        so there is one less column
    tempest_extremes_header_str : str, default="start"
        This is an option in the Colin's load function, so I assume this can change
        between files

    track_calendar : str, optional
          When loading data from a TRACK ASCII file, if the data uses a different
          calendar to the default :class:`datetime.datetime`, then you can pass this
          argument to load the times in as :class:`cftime.datetime` with the given
          calendar instead
          Can also be set to "timestep" if the TRACK file has timesteps instead of dates.
          In that case, you need to provide initial_date and timestep arguments.
    **kwargs
        When loading tracks from a standard files these will be passed to the relevant
        load function

        * netCDF file - :func:`xarray.open_dataset`
        * CSV file - :func:`pandas.read_csv`
        * parquet file - :func:`pandas.read_parquet`

        For CSV files pandas interprets "NA" as `nan` by default, which is overriden in
        this function. To restore the pandas default behavious set
        :code:`keep_default_NA=True` and :code:`na_values=[]`

    Returns
    -------
    xarray.Dataset

    """
    # Overwrite default arguments with explicit arguments passed to rename by putting
    # "rename" second in this dictionary combination
    rename = {**rename_defaults, **rename}

    # If source is not given, try to derive the right function from the file extension
    if source is None:
        extension = filename.split(".")[-1]
        if extension == "csv":
            data = _csv.load(filename, **kwargs)
        elif extension == "parquet":
            data = _csv.load(filename, load_function=pd.read_parquet, **kwargs)
        elif filename.split(".")[-1] == "nc":
            data = _netcdf.load(filename, rename, **kwargs)
        else:
            raise ValueError("Source is set to None and file type is not detected")

    # If source is given, use the relevant function
    else:
        source = source.lower()
        if source == "track":
            data = _TRACK.load(
                filename,
                calendar=track_calendar,
                variable_names=variable_names,
                **kwargs,
            )
        elif source == "track.tilt":
            data = _TRACK.load_tilts(
                filename,
                calendar=track_calendar,
            )
        elif source in ["csv", "uz"]:
            data = _csv.load(filename, **kwargs)
        elif source in ["te", "tempest", "tempestextremes"]:
            data = _tempestextremes.load(
                filename,
                variable_names,
                tempest_extremes_unstructured,
                tempest_extremes_header_str,
            )
        elif source == "witrack":
            data = witrack.load(filename)
        elif source == "ibtracs":
            data = ibtracs.load(ibtracs_subset, filename, **kwargs)
        elif source == "netcdf":
            data = _netcdf.load(filename, rename, **kwargs)
        elif source in [
            "old_hurdat",
            "ecmwf",
        ]:
            data = _old_HURDAT.load(filename)
        else:
            raise ValueError(f"Source {source} unsupported or misspelled")

    # xarray.Dataset.rename only accepts keys that are actually in the dataset
    rename = {key: rename[key] for key in rename if key in data}

    if len(rename) > 0:
        data = data.rename(rename)

    if units is not None:
        for varname in units:
            data[varname].attrs["units"] = units[varname]

    if baselon is not None:
        data["lon"] = ((data.lon - baselon) % 360) + baselon

    return data



import numpy as np
from metpy.constants import Re, omega
from metpy.units import units
from metpy.xarray import preprocess_and_wrap
import pint

from .._metpy import dequantify_results


omega = omega / units("radian")


@dequantify_results
@preprocess_and_wrap(wrap_like=("lat", "lat"))
def beta_drift(
    lat,
    wind_max,
    radius_wind_max,
):
    """
    Based on Smith, 1993: https://journals.ametsoc.org/view/journals/atsc/50/18/1520-0469_1993_050_3213_ahbdl_2_0_co_2.xml?tab_body=pdf

    Parameters
    ----------
    lat : TYPE
        DESCRIPTION.
    wind_max : TYPE
        DESCRIPTION.
    radius_wind_max : TYPE
        DESCRIPTION.

    Returns
    -------
    V_drift : TYPE
        DESCRIPTION.
    theta_drift : TYPE
        DESCRIPTION.

    """

    # Treat input
    ## Convert lat to rad
    if not isinstance(lat, pint.Quantity) or lat.unitless:
        if np.abs(lat).max() > np.pi:  # We assume lats are in degrees if they exceed pi
            lat = lat * units("degrees")
        else:
            lat = lat * units("radians")

    lat = lat.to("radians")

    # Assume max wind is in m/s if not given
    if not isinstance(wind_max, pint.Quantity) or wind_max.unitless:
        wind_max = wind_max * units("m / s")

    ## Convert rmw to m
    if not isinstance(radius_wind_max, pint.Quantity) or radius_wind_max.unitless:
        if (
            radius_wind_max.max() < 10000
        ):  # We assume rmw are in km if they are below 10,000
            radius_wind_max = radius_wind_max * 1000

        radius_wind_max = radius_wind_max * units("m")

    # Coriolis parameter
    beta = 2 * omega * np.cos(lat) / Re  # s-1 m-1

    # Beta-drift parameters
    V_char = (radius_wind_max**2) * beta  # m/s
    B = V_char / wind_max  # non-dimensionnal
    V_drift_adim = 0.72 * B ** (-0.54)  # non-dmensionnal
    V_drift = V_drift_adim * V_char  # m/s
    theta_drift = 308 - 9.6 * np.log(B)  # degrees

    return V_drift, theta_drift
